﻿using System;
using System.Collections.Generic;
using System.Linq;
using MVR.FileManagementSecure;
using SimpleJSON;
using UnityEngine;

namespace Evo {
    /// <summary>
    ///     PoseConvert v2 by Evo
    ///     Load a rig from a saved pose to the current Person Atom without setting control positions.
    /// </summary>
    public class PoseConvert : MVRScript {
        private static readonly string PosesPath = "Custom/Atom/Person/Pose";

        private static readonly string[] PersonAtomControlNames = {
            "hipControl", "pelvisControl",

            "rThighControl", "rKneeControl", "rFootControl", "rToeControl",

            "lThighControl", "lKneeControl", "lFootControl", "lToeControl",

            "abdomenControl", "abdomen2Control", "chestControl",

            "rShoulderControl", "rArmControl", "rElbowControl", "rHandControl",

            "lShoulderControl", "lArmControl", "lElbowControl", "lHandControl",

            "neckControl", "headControl",

            "eyeTargetControl"
        };

        private static readonly string[] MaleGensControlNames = {
            "penisBaseControl", "penisMidControl", "penisTipControl", "testesControl"
        };

        private static readonly string[] ControlOptionFloatParams = {
            "complyPositionThreshold", "complyRotationThreshold", "complySpeed", "complyJointDriveSpring",
            "complyPositionSpring", "complyRotationSpring", "complyPositionDamper", "complyRotationDamper",

            "positionGrid", "rotationGrid", "mass", "drag", "maxVelocity", "angularDrag",

            "holdPositionSpring", "holdPositionDamper", "holdPositionMaxForce", "holdRotationSpring",
            "holdRotationDamper",
            "holdRotationMaxForce",

            "linkPositionSpring", "linkPositionDamper", "linkPositionMaxForce", "linkRotationSpring",
            "linkRotationDamper",
            "linkRotationMaxForce",

            "jointDriveSpring", "jointDriveDamper", "jointDriveMaxForce", "jointDriveXTarget", "jointDriveYTarget",
            "jointDriveZTarget"
        };

        private UIDynamicButton applyLoadedRigToTargetBtn;

        private List<string> controlPresetChoices;
        private JSONStorableStringChooser controlPresetChooser;

        private JSONStorableBool copyFloatParams;
        private JSONStorableBool copyMaleGensControls;
        private JSONStorableBool copyParentLinks;

        private List<JSONStorableBool> forcedPositionStateBools;
        private JSONStorableBool forceElbowsPositionOn;
        private JSONStorableBool forceElbowsRotationOn;
        private JSONStorableBool forceFeetPositionOn;
        private JSONStorableBool forceFeetRotationOn;
        private JSONStorableBool forceHandsPositionOn;
        private JSONStorableBool forceHandsRotationOn;
        private JSONStorableBool forceHeadPositionOn;
        private JSONStorableBool forceHeadRotationOn;
        private JSONStorableBool forceKneesPositionOn;
        private JSONStorableBool forceKneesRotationOn;

        private JSONStorableBool forceNeckPositionOn;
        private JSONStorableBool forceNeckRotationOn;

        private bool isMale;
        private List<JSONNode> loadedRigStorables;
        private UIDynamicTextField loadedRigTextField;

        private UIDynamicButton loadRigFromPoseBtn;

        private Atom targetPersonAtom;

        public override void Init() {
            targetPersonAtom = containingAtom;
            loadedRigStorables = new List<JSONNode>();
            controlPresetChoices = new List<string> {
                "Inherit", "All", "Position All", "Rotation All", "Extremities", "Position Extremities",
                "Rotation Extremities"
            };
            var dazCharacterSelector = targetPersonAtom.GetStorableByID("geometry") as DAZCharacterSelector;
            isMale = dazCharacterSelector.selectedCharacter.isMale;

            buildUI();
        }

        private void buildUI() {
            CreateTextField(new JSONStorableString("LeftColumn"
                    , "<size=40><b>Load Rig From Pose</b></size>\n" +
                      "Copy rig properties from another pose without setting the positions of the current pose."))
                .height = 150f;

            loadRigFromPoseBtn = CreateButton("Load Source Rig From Pose");
            loadRigFromPoseBtn.button.onClick.AddListener(loadSourcePoseRig);
            RegisterAction(new JSONStorableAction("LoadRigFromPose", loadSourcePoseRig));

            loadedRigTextField = CreateTextField(
                new JSONStorableString("LoadedRigNameLabel",
                    ""), true);
            loadedRigTextField.height = 150f;

            applyLoadedRigToTargetBtn = CreateButton("No Rig Loaded", true);
            applyLoadedRigToTargetBtn.buttonColor = Color.red;
            applyLoadedRigToTargetBtn.button.enabled = false;
            applyLoadedRigToTargetBtn.button.onClick.AddListener(applyLoadedRigToTarget);
            RegisterAction(new JSONStorableAction("ApplyLoadedRigToTarget", applyLoadedRigToTarget));

            CreateSpacer().height = 45f;
            CreateSpacer(true).height = 25f;

            copyFloatParams = new JSONStorableBool("CopyControlOptions", true);
            copyParentLinks = new JSONStorableBool("CopyParentLinking", true);
            copyMaleGensControls = new JSONStorableBool("CopyMaleGensControls", false);

            CreateToggle(copyFloatParams).label = "Copy Float Params";
            CreateToggle(copyParentLinks).label = "Copy Parent Links";
            if (isMale)
                CreateToggle(copyMaleGensControls).label = "Copy Male Gens Controls";
            else
                CreateSpacer().height = 50f;

            CreateSpacer().height = 10f;
            CreateSpacer(true).height = 240f;

            controlPresetChooser = new JSONStorableStringChooser("Control Preset", controlPresetChoices,
                controlPresetChoices.First(),
                "Control Preset",
                applyPositionModePreset);
            RegisterStringChooser(controlPresetChooser);
            CreatePopup(controlPresetChooser, true);

            CreateSpacer(true)
                .height = 10f;

            CreateTextField(new JSONStorableString("LeftColumn"
                    , "<size=40><b>Individual Controls</b></size>\n" +
                      "Override the position mode of controls of the source pose."))
                .height = 120f;

            forcedPositionStateBools = new List<JSONStorableBool>();

            forceNeckPositionOn = new JSONStorableBool("ForceNeckPositionOn", false);
            forcedPositionStateBools.Add(forceNeckPositionOn);
            forceNeckRotationOn = new JSONStorableBool("ForceNeckRotationOn", false);
            forcedPositionStateBools.Add(forceNeckRotationOn);
            forceHeadPositionOn = new JSONStorableBool("ForceHeadPositionOn", false);
            forcedPositionStateBools.Add(forceHeadPositionOn);
            forceHeadRotationOn = new JSONStorableBool("ForceHeadRotationOn", false);
            forcedPositionStateBools.Add(forceHeadRotationOn);
            forceHandsPositionOn = new JSONStorableBool("ForceHandsPositionOn", false);
            forcedPositionStateBools.Add(forceHandsPositionOn);
            forceHandsRotationOn = new JSONStorableBool("ForceHandsRotationOn", false);
            forcedPositionStateBools.Add(forceHandsRotationOn);
            forceElbowsPositionOn = new JSONStorableBool("ForceElbowsPositionOn", false);
            forcedPositionStateBools.Add(forceElbowsPositionOn);
            forceElbowsRotationOn = new JSONStorableBool("ForceElbowsRotationOn", false);
            forcedPositionStateBools.Add(forceElbowsRotationOn);
            forceKneesPositionOn = new JSONStorableBool("ForceKneesPositionOn", false);
            forcedPositionStateBools.Add(forceKneesPositionOn);
            forceKneesRotationOn = new JSONStorableBool("ForceKneesRotationOn", false);
            forcedPositionStateBools.Add(forceKneesRotationOn);
            forceFeetPositionOn = new JSONStorableBool("ForceFootPositionOn", false);
            forcedPositionStateBools.Add(forceFeetPositionOn);
            forceFeetRotationOn = new JSONStorableBool("ForceFootRotationOn", false);
            forcedPositionStateBools.Add(forceFeetRotationOn);

            CreateToggle(forceNeckPositionOn).label = "Neck Position On";
            CreateToggle(forceNeckRotationOn, true).label = "Neck Rotation On";
            CreateToggle(forceHeadPositionOn).label = "Head Position On";
            CreateToggle(forceHeadRotationOn, true).label = "Head Rotation On";
            CreateToggle(forceHandsPositionOn).label = "Hand Position On";
            CreateToggle(forceHandsRotationOn, true).label = "Hand Rotation On";
            CreateToggle(forceElbowsPositionOn).label = "Elbow Position On";
            CreateToggle(forceElbowsRotationOn, true).label = "Elbow Rotation On";
            CreateToggle(forceKneesPositionOn).label = "Knee Position On";
            CreateToggle(forceKneesRotationOn, true).label = "Knee Rotation On";
            CreateToggle(forceFeetPositionOn).label = "Foot Position On";
            CreateToggle(forceFeetRotationOn, true).label = "Foot Rotation On";
        }

        private void loadSourcePoseRig() {
            var vapPoseShortcuts =
                FileManagerSecure.GetShortCutsForDirectory(PosesPath, false, false, true, true);
            SuperController.singleton.GetMediaPathDialog(handlePoseSelected, "vap", PosesPath, false, true,
                false, null, false, vapPoseShortcuts);
        }

        private void handlePoseSelected(string path) {
            if (!string.IsNullOrEmpty(path)) {
                var poseJson = JSON.Parse(SuperController.singleton.ReadFileIntoString(path));
                var controlNames = PersonAtomControlNames.Concat(MaleGensControlNames);

                loadedRigStorables = new List<JSONNode>();
                foreach (JSONNode node in poseJson["storables"].AsArray)
                    if (controlNames.Contains(node["id"].Value))
                        loadedRigStorables.Add(node);

                applyLoadedRigToTargetBtn.button.enabled = true;
                applyLoadedRigToTargetBtn.label = "Apply Rig to Current Pose";
                applyLoadedRigToTargetBtn.buttonColor = Color.green;

                loadedRigTextField.text = path;
            }
        }

        private void applyLoadedRigToTarget() {
            var bodyControls = targetPersonAtom.freeControllers.Where(control =>
                PersonAtomControlNames.Contains(control.name));

            if (copyMaleGensControls.val) {
                var maleGensControls = targetPersonAtom.freeControllers.Where(control => 
                    MaleGensControlNames.Contains(control.name));
                bodyControls = bodyControls.Concat(maleGensControls);
            }

            foreach (var control in bodyControls) {
                var node = loadedRigStorables.Find(it => it["id"].Value == control.name);

                if (node == null) {
                    SuperController.LogMessage($"Pose file does not contain a value for: {control.name}, skipping.");
                    continue;
                }

                if (copyParentLinks.val) {
                    var linkToControlNameRaw = node["linkTo"]?.Value;

                    if (!string.IsNullOrEmpty(linkToControlNameRaw)) {
                        var localPersonControlTarget = linkToControlNameRaw.Split(':').Skip(1).First();
                        var linkToAtomName = $"{targetPersonAtom.uid}:{localPersonControlTarget}";

                        var linkToRb = SuperController.singleton.RigidbodyNameToRigidbody(linkToAtomName);
                        control.SelectLinkToRigidbody(linkToRb, FreeControllerV3.SelectLinkState.PositionAndRotation);
                    }
                    else {
                        // Unlink any current
                        var noneRb = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                        control.SelectLinkToRigidbody(noneRb, FreeControllerV3.SelectLinkState.PositionAndRotation);
                    }
                }

                control.canGrabPosition = node["canGrabPosition"].AsBool;
                control.canGrabRotation = node["canGrabRotation"].AsBool;

                var positionState = (FreeControllerV3.PositionState)Enum.Parse(
                    typeof(FreeControllerV3.PositionState), node["positionState"].Value);
                control.currentPositionState = positionState;
                var rotationState = (FreeControllerV3.RotationState)Enum.Parse(
                    typeof(FreeControllerV3.RotationState), node["rotationState"].Value);
                control.currentRotationState = rotationState;

                var controlNameLower = control.name.ToLower();

                if (controlNameLower.Contains("neck")) {
                    if (forceNeckPositionOn.val)
                        control.currentPositionState = FreeControllerV3.PositionState.On;
                    if (forceNeckRotationOn.val)
                        control.currentRotationState = FreeControllerV3.RotationState.On;
                }

                if (controlNameLower.Contains("head")) {
                    if (forceHeadPositionOn.val)
                        control.currentPositionState = FreeControllerV3.PositionState.On;
                    if (forceHeadRotationOn.val)
                        control.currentRotationState = FreeControllerV3.RotationState.On;
                }

                if (controlNameLower.Contains("hand")) {
                    if (forceHandsPositionOn.val)
                        control.currentPositionState = FreeControllerV3.PositionState.On;
                    if (forceHandsRotationOn.val)
                        control.currentRotationState = FreeControllerV3.RotationState.On;
                }

                if (controlNameLower.Contains("elbow")) {
                    if (forceElbowsPositionOn.val)
                        control.currentPositionState = FreeControllerV3.PositionState.On;
                    if (forceElbowsRotationOn.val)
                        control.currentRotationState = FreeControllerV3.RotationState.On;
                }

                if (controlNameLower.Contains("knee")) {
                    if (forceKneesPositionOn.val)
                        control.currentPositionState = FreeControllerV3.PositionState.On;
                    if (forceKneesRotationOn.val)
                        control.currentRotationState = FreeControllerV3.RotationState.On;
                }

                if (controlNameLower.Contains("foot")) {
                    if (forceFeetPositionOn.val)
                        control.currentPositionState = FreeControllerV3.PositionState.On;
                    if (forceFeetRotationOn.val)
                        control.currentRotationState = FreeControllerV3.RotationState.On;
                }

                if (copyFloatParams.val)
                    foreach (var floatParam in ControlOptionFloatParams)
                        control.SetFloatParamValue(floatParam, node[floatParam].AsFloat);
            }
        }

        private bool isExtremity(JSONStorableBool forcedState) {
            return forcedState.name.Contains("Hand") ||
                   forcedState.name.Contains("Foot");
        }

        private void applyPositionModePreset(string presetName) {
            foreach (var forcedPositionState in forcedPositionStateBools)
                forcedPositionState.val = false;

            switch (presetName) {
                case "Inherit":
                    foreach (var forcedState in forcedPositionStateBools)
                        forcedState.val = false;
                    break;
                case "All":
                    foreach (var forcedPositionState in forcedPositionStateBools)
                        forcedPositionState.val = true;
                    break;
                case "Position All":
                    foreach (var forcedState in forcedPositionStateBools)
                        if (forcedState.name.Contains("Position"))
                            forcedState.val = true;
                    break;
                case "Rotation All":
                    foreach (var forcedState in forcedPositionStateBools)
                        if (forcedState.name.Contains("Rotation"))
                            forcedState.val = true;
                    break;
                case "Extremities":
                    foreach (var forcedState in forcedPositionStateBools)
                        if (isExtremity(forcedState))
                            forcedState.val = true;
                    break;
                case "Position Extremities":
                    foreach (var forcedState in forcedPositionStateBools)
                        if (forcedState.name.Contains("Position") && isExtremity(forcedState))
                            forcedState.val = true;
                    break;
                case "Rotation Extremities":
                    foreach (var forcedState in forcedPositionStateBools)
                        if (forcedState.name.Contains("Rotation") && isExtremity(forcedState))
                            forcedState.val = true;
                    break;
                default:
                    throw new Exception("Invalid Position Override Preset Name.");
            }
        }
    }
}